==> webform_handler_defer-1.1.0.tar.gz: webform_handler_defer/src/WebformDeferredProcessor.php <==
<?php

namespace Drupal\webform_handler_defer;

use Drupal\user\Entity\User;
use Drupal\webform\Entity\Webform;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\Queue\QueueInterface;
use Drupal\webform\Entity\WebformSubmission;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Session\AccountSwitcherInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;

/**
 * Provides a deferral and deferred processing of webform submission handlers.
 */
class WebformDeferredProcessor {

  /**
   * Logger channel.
   */
  protected LoggerChannelInterface $logger;

  /**
   * Queue service.
   */
  protected QueueInterface $queue;

  /**
   * Account Switcher service.
   */
  protected AccountSwitcherInterface $accountSwitcher;

  /**
   * The active Webform.
   */
  protected Webform $webform;

  /**
   * The active Webform submission.
   */
  protected WebformSubmission $submission;

  /**
   * Array of IDs of handlers to be deferred.
   */
  protected array $deferredHandlers;

  /**
   * WebformDeferredProcessor constructor.
   *
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $loggerFactory
   *   The logger channel factory service.
   * @param \Drupal\Core\Queue\QueueFactory $queueFactory
   *   The queue service.
   * @param \Drupal\Core\Session\AccountSwitcherInterface $accountSwitcher
   *   The account switcher service.
   */
  public function __construct(
    LoggerChannelFactoryInterface $loggerFactory,
    QueueFactory $queueFactory,
    AccountSwitcherInterface $accountSwitcher,
  ) {
    $this->logger = $loggerFactory->get('webform_handler_defer');
    $this->queue = $queueFactory->get('webform_handler_defer_submissions');
    $this->accountSwitcher = $accountSwitcher;
  }

  /**
   * Log submission ID and disable submission handlers.
   *
   * @return bool
   *   Success or failure.
   */
  public function defer() : bool {
    $record = [
      'sid' => $this->submission->id(),
      'deferred' => $this->deferredHandlers,
    ];

    // Add item to queue, check for and return failure.
    if ($this->queue->createItem($record) === FALSE) {
      return FALSE;
    }

    // Disable all deferred handlers and return success.
    $this->disableHandlers(TRUE);
    return TRUE;
  }

  /**
   * Cron wrapper: process deferred handlers for previous submissions.
   */
  public function cron() {
    $this->processSubmissions();
  }

  /**
   * Process the submission handlers which were deferred.
   *
   * Our strategy for handling any kind of failed processing here:
   * Aim to leave things *as they would have been* if the handlers had not
   * been deferred, but processed as normal during the original submission.
   */
  protected function processSubmissions() {
    // Process items in the queue.
    $submissions_count = 0;
    while (!empty($item = $this->queue->claimItem())) {
      $record = $item->data;

      try {
        // Delete from the queue immediately, to prevent it getting stuck
        // or repeated processing due to some unexpected error condition.
        $this->queue->deleteItem($item);

        $submissions_count++;
        if (empty($submission = WebformSubmission::load($record['sid']))) {
          // Submission has been deleted, user does not have access, etc.
          continue;
        }
        $this->setup($submission, $record['deferred'])->sudoRunHandlers();
      }
      catch (\Exception $e) {
        $this->logger->error('Error during deferred processing of submission @submission_id: @error', [
          '@submission_id' => $record['sid'],
          '@error' => $e->getMessage(),
        ]);
      }
    }

    // Log a summary of what was done.
    $this->logger->info(sprintf('Deferred handlers for %d submissions were processed.', $submissions_count));
  }

  /**
   * Switch user as needed, then run the deferred handlers.
   *
   * @return object
   *   $this
   */
  protected function sudoRunHandlers() : object {
    $user_anon = User::load(0);
    $user_admin = User::load(1);
    $user_current = \Drupal::currentUser();

    // Become admin user so that we can reliably access the submission.
    $this->accountSwitcher->switchTo($user_admin);
    // Identify the user who made the submission. Fall back to anonymous.
    $submitter = $this->submission->getOwner() ?? $user_anon;
    $this->accountSwitcher->switchBack();

    // Run handlers using the original user account.
    if ($user_current->id() == $submitter->id()) {
      $this->invokeDeferredHandlers();
    }
    else {
      try {
        $this->accountSwitcher->switchTo($submitter);
        $this->invokeDeferredHandlers();
      }
      finally {
        $this->accountSwitcher->switchBack();
      }
    }

    return $this;
  }

  /**
   * Execute the deferred handlers for the given submission.
   *
   * @return object
   *   $this
   */
  protected function invokeDeferredHandlers() {

    // Disable all handlers *except* the deferred handlers.
    $this->disableHandlers(FALSE);
    // Invoke the remaining handlers.
    $this->webform->invokeHandlers('postSave', $this->submission);

    return $this;
  }

  /**
   * Disable selected submission handlers.
   *
   * @param bool $deferred
   *   Whether to disable the deferred or non-deferred handlers.
   *
   * @return object
   *   $this
   */
  protected function disableHandlers($deferred) {

    foreach ($this->webform->getHandlers() as $handler_id => $handler) {
      if (in_array($handler_id, $this->deferredHandlers, TRUE) == $deferred) {
        $handler->disable();
      }
    }

    return $this;
  }

  /**
   * Set up processing of a submission.
   *
   * @param \Drupal\webform\Entity\WebformSubmissionInterface $submission
   *   Webform submission object.
   * @param array $defer_handlers
   *   IDs of handlers to be deferred.
   *
   * @return object
   *   $this
   */
  public function setup(WebformSubmission $submission, array $defer_handlers) : object {
    $this->submission = $submission;
    $this->webform = $submission->getWebform();
    $this->deferredHandlers = $defer_handlers;
    return $this;
  }

}

==> webform_handler_defer-1.1.1.tar.gz: webform_handler_defer/src/WebformDeferredProcessor.php <==
<?php

namespace Drupal\webform_handler_defer;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\Queue\QueueInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\Session\AccountSwitcherInterface;
use Drupal\webform\Entity\Webform;
use Drupal\webform\Entity\WebformSubmission;

/**
 * Provides a deferral and deferred processing of webform submission handlers.
 */
class WebformDeferredProcessor {

  /**
   * Logger channel.
   */
  protected LoggerChannelInterface $logger;

  /**
   * Queue service.
   */
  protected QueueInterface $queue;

  /**
   * Current user account.
   */
  protected AccountProxyInterface $currentUser;

  /**
   * Account Switcher service.
   */
  protected AccountSwitcherInterface $accountSwitcher;

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The active Webform.
   */
  protected Webform $webform;

  /**
   * The active Webform submission.
   */
  protected WebformSubmission $submission;

  /**
   * Array of IDs of handlers to be deferred.
   */
  protected array $deferredHandlers;

  /**
   * WebformDeferredProcessor constructor.
   *
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $loggerFactory
   *   The logger channel factory service.
   * @param \Drupal\Core\Queue\QueueFactory $queueFactory
   *   The queue service.
   * @param \Drupal\Core\Session\AccountProxyInterface $current_user
   *   The current user service.
   * @param \Drupal\Core\Session\AccountSwitcherInterface $accountSwitcher
   *   The account switcher service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   */
  public function __construct(
    LoggerChannelFactoryInterface $loggerFactory,
    QueueFactory $queueFactory,
    AccountProxyInterface $current_user,
    AccountSwitcherInterface $accountSwitcher,
    EntityTypeManagerInterface $entity_type_manager,
  ) {
    $this->logger = $loggerFactory->get('webform_handler_defer');
    $this->queue = $queueFactory->get('webform_handler_defer_submissions');
    $this->currentUser = $current_user;
    $this->accountSwitcher = $accountSwitcher;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * Log submission ID and disable submission handlers.
   *
   * @return bool
   *   Success or failure.
   */
  public function defer() : bool {
    $record = [
      'sid' => $this->submission->id(),
      'deferred' => $this->deferredHandlers,
    ];

    // Add item to queue, check for and return failure.
    if ($this->queue->createItem($record) === FALSE) {
      return FALSE;
    }

    // Disable all deferred handlers and return success.
    $this->disableHandlers(TRUE);
    return TRUE;
  }

  /**
   * Cron wrapper: process deferred handlers for previous submissions.
   */
  public function cron() {
    $this->processSubmissions();
  }

  /**
   * Process the submission handlers which were deferred.
   *
   * Our strategy for handling any kind of failed processing here:
   * Aim to leave things *as they would have been* if the handlers had not
   * been deferred, but processed as normal during the original submission.
   */
  protected function processSubmissions() {
    // Process items in the queue.
    $submissions_count = 0;
    while (!empty($item = $this->queue->claimItem())) {
      $record = $item->data;

      try {
        // Delete from the queue immediately, to prevent it getting stuck
        // or repeated processing due to some unexpected error condition.
        $this->queue->deleteItem($item);

        $submissions_count++;
        if (empty($submission = WebformSubmission::load($record['sid']))) {
          // Submission has been deleted, user does not have access, etc.
          continue;
        }
        $this->setup($submission, $record['deferred'])->sudoRunHandlers();
      }
      catch (\Exception $e) {
        $this->logger->error('Error during deferred processing of submission @submission_id: @error', [
          '@submission_id' => $record['sid'],
          '@error' => $e->getMessage(),
        ]);
      }
    }

    // Log a summary of what was done.
    $this->logger->info(sprintf('Deferred handlers for %d submissions were processed.', $submissions_count));
  }

  /**
   * Switch user as needed, then run the deferred handlers.
   *
   * @return object
   *   $this
   */
  protected function sudoRunHandlers() : object {
    // Get anonymous and admin user accounts.
    $user_storage = $this->entityTypeManager->getStorage('user');
    $anon_user = $user_storage->load(0);
    $admin_user = $user_storage->load(1);

    // Become admin user so that we can reliably access the submission.
    $this->accountSwitcher->switchTo($admin_user);
    // Identify the user who made the submission. Fall back to anonymous.
    $submitter = $this->submission->getOwner() ?? $anon_user;
    $this->accountSwitcher->switchBack();

    // Run handlers using the original user account.
    if ($this->currentUser->id() == $submitter->id()) {
      $this->invokeDeferredHandlers();
    }
    else {
      try {
        $this->accountSwitcher->switchTo($submitter);
        $this->invokeDeferredHandlers();
      }
      finally {
        $this->accountSwitcher->switchBack();
      }
    }

    return $this;
  }

  /**
   * Execute the deferred handlers for the given submission.
   *
   * @return object
   *   $this
   */
  protected function invokeDeferredHandlers() {

    // Disable all handlers *except* the deferred handlers.
    $this->disableHandlers(FALSE);
    // Invoke the remaining handlers.
    $this->webform->invokeHandlers('postSave', $this->submission);

    return $this;
  }

  /**
   * Disable selected submission handlers.
   *
   * @param bool $deferred
   *   Whether to disable the deferred or non-deferred handlers.
   *
   * @return object
   *   $this
   */
  protected function disableHandlers($deferred) {

    foreach ($this->webform->getHandlers() as $handler_id => $handler) {
      if (in_array($handler_id, $this->deferredHandlers, TRUE) == $deferred) {
        $handler->disable();
      }
    }

    return $this;
  }

  /**
   * Set up processing of a submission.
   *
   * @param \Drupal\webform\Entity\WebformSubmissionInterface $submission
   *   Webform submission object.
   * @param array $defer_handlers
   *   IDs of handlers to be deferred.
   *
   * @return object
   *   $this
   */
  public function setup(WebformSubmission $submission, array $defer_handlers) : object {
    $this->submission = $submission;
    $this->webform = $submission->getWebform();
    $this->deferredHandlers = $defer_handlers;
    return $this;
  }

}
